unit UnitFrmMouseMonitor;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, Menus, UnitFrmDummyUnicodeTooltip, UnitFrmRatTrap;

type TMouseDirection = ( mdUp, mdDown, mdLeft, mdRight, mdNE, mdNW, mdSE, mdSW,  mdNone);
type
  TFrmMouseMonitor = class(TForm)
    timMouse: TTimer;
    txtXavg: TStaticText;
    txtYavg: TStaticText;
    MemoLog: TMemo;
    txtUp: TStaticText;
    txtDown: TStaticText;
    txtLeft: TStaticText;
    txtRight: TStaticText;
    txtDirection: TStaticText;
    pmDummy: TPopupMenu;
    a1: TMenuItem;
    timPaused: TTimer;
    imgDown: TImage;
    imgLeft: TImage;
    imgUp: TImage;
    imgRight: TImage;
    txtStartX: TStaticText;
    txtStartY: TStaticText;
    txtmX: TStaticText;
    txtmY: TStaticText;
    txtPercent: TStaticText;
    imgNE: TImage;
    imgNW: TImage;
    imgSE: TImage;
    imgSW: TImage;
    procedure timMouseTimer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure timPausedTimer(Sender: TObject);
  private
    { Private declarations }
        // INI settings
        xdelta, ydelta : integer;
        deltaBias : single;
        xThreshold, yThreshold, timethreshold : integer;
        XPixelThreshold : integer;
        YPixelThreshold : integer;
        directionBias : single;
        helpPause : integer;
        directionThreshold : integer;
        EnableHelpPause : boolean;
        EnableDiagonals : boolean;
        EnableDrawing : boolean;

        // other settings
        lastx, lasty : integer;
        interpreted : boolean;
        timeheld : integer;
        xavg, yavg : integer;
        xbinary, ybinary : integer;
        xSwapped, ySwapped : boolean;

        storedX, storedY : integer;

        lastClock : cardinal;
        leftavg, rightavg, upavg, downavg : integer;
        startingx, startingy : integer;
        angleavg : integer;
        peravgmax : integer;
        //angleavg : integer;
        leftcount, rightcount, upcount, downcount : integer;
        NEcount, NWcount, SEcount, SWcount : integer;
        NEavg, NWavg, SEavg, SWavg : integer;
        detectedDirection : TMouseDirection;
        vertDirection, horzDirection : TMouseDirection;
        westDirection, eastDirection : TMouseDirection;
        vert, horz : TMouseDirection;
        eastquad, westquad : TMouseDirection;
        lastDetectedDirection : TMouseDirection;
        lastAddDirection : TMouseDirection;
        LastAddActual : TMouseDirection;
        LastChangeDirection : TMouseDirection;
        lastlen : integer;
        lastangle : integer;
        directionlist : TList;
        directionLen :Tlist;
        directionAngle : TList;
        fullscreen : boolean;
        firstdirection : TMouseDirection;
        dupcount : integer;
        abort : boolean;
        disabled : boolean;
        lastLogLin : string;
        tt : TTooltipWindow;
        LastTooltipText : string;
        function IsFullScreen : boolean;


    procedure SetHelpPause(const Value: integer);

  public
    { Public declarations }
        function ToString(md : TMouseDirection) : string;
        function IsDiagonal(md : TMouseDirection) : boolean;
        function IsNotDiagonal(md : TMouseDirection) : boolean;
        procedure AbortInterpret;
        procedure SetMonitorEnabled(enable : boolean);

        procedure GetGesturesSuggestion(gesture : string; suggest : TStringList);
        procedure ClosePopups;

        function GetDirectionAngle(index : integer) : integer;
        function IsBetweenNeighbors(dir1, dir2 : TMouseDirection; angle : integer) : boolean;

        procedure AppendLog(line : string);

        property prDirectionBias : single read DirectionBias write DirectionBias;
        property prDeltaBias : single read deltaBias write deltaBias;
        property prYPixelThreshold : integer read YPixelThreshold write YPixelThreshold;
        property prXPixelThreshold : integer read XPixelThreshold write XPixelThreshold;
        property prXdelta : integer read xdelta write xdelta;
        property prYdelta : integer read ydelta write ydelta;
        property prXThreshold : integer read xThreshold write xThreshold;
        property prYThreshold : integer read yThreshold write yThreshold;
        property prTimethreshold : integer read timethreshold write timethreshold;
        property prHelpPause : integer read HelpPause write SetHelpPause;
        property prDirectionThreshold : integer read DirectionThreshold write DirectionThreshold;
        property prEnableHelpPause : boolean read EnableHelpPause write EnableHelpPause;
        property prEnableDiagonals : boolean read EnableDiagonals write EnableDiagonals;
        property prEnableDrawing : Boolean read EnableDrawing write EnableDrawing;
        property prStartingX : Integer read startingx;
        property prStartingY : Integer read startingy;
  end;

var
  FrmMouseMonitor: TFrmMouseMonitor;

implementation

uses UnitKeyboardQuery, Math, IniFiles, UnitMisc, UnitFrmGestures, UnitPaste,
  UnitFrmScratchpad, UnitFrmLines;

{$R *.dfm}


const RIGHT_A = 0;
            DOWN_A = 90;
            LEFT_A = 180;
            UP_A = 270;
            NE_A = UP_A + 45;
            NW_A = UP_A - 45;
            SW_A = DOWN_A + 45;
            SE_A = DOWN_A - 45;

procedure TFrmMouseMonitor.AbortInterpret;
begin
    self.abort := true;
end;

procedure TFrmMouseMonitor.AppendLog(line: string);
begin
    if (line <> lastLogLin)  then begin
        if dupcount <> 0 then begin
            MemoLog.Lines.Strings[ memolog.lines.Count - 1] := MemoLog.Lines.Strings[ memolog.lines.Count - 1] + '(*' + IntToSTr(dupcount)+')';
        end;
        Memolog.Lines.add(line);
        dupcount := 0;
    end else begin

        inc(dupcount);
    end;
    lastLogLin := line;
end;

procedure TFrmMouseMonitor.ClosePopups;
const WM_EXITMENULOOP = $0212;
var    h : THandle;
begin

    pmDummy.Free;
    pmDummy := TPopupMenu.Create(FrmMouseMonitor);
    h := Windows.GetForegroundWindow;

    if threadattach(h) then begin
        Windows.SetFocus(pmDummy.WindowHandle); // focus require
        pmDummy.Popup(0,0);
        //Windows.SetFocus(h);
        ThreadDetach;
    end;


    //Windows.SendMessage(pmDummy.WindowHandle,WM_EXITMENULOOP,0,0);
    //Windows.PostMessage(self.Handle, WM_NULL, 0, 0);
end;

procedure TFrmMouseMonitor.FormCreate(Sender: TObject);
begin
    timPaused.Interval := helpPause;
    directionlist := TList.Create;
    directionLen := TList.Create;
    directionAngle := TList.Create;
    tt := TTooltipWindow.Create;
end;

function TFrmMouseMonitor.GetDirectionAngle(index: integer): integer;
begin
    result := Integer(directionAngle.Items[index]);
end;

function IsBetween(dir1, dir2 : TMouseDirection; angle : integer) : boolean;
begin

end;

procedure TFrmMouseMonitor.GetGesturesSuggestion(gesture: string;
  suggest: TStringList);
    function ErrorPos(dir1, dir2 : TMouseDirection) : integer;
    begin
        result := Pos( IntToStr(Integer(dir1)) + IntToStr(Integer(dir2)), gesture );
    end;

    function GetExtrapAngle(dir1, dir2 : TMouseDirection) : integer;
    var err : integer;
        angle1, len1, angle2, len2 : integer;
    begin
        err := ErrorPos(dir1, dir2);
        dec(err);
        result := err;
        if err <> -1 then begin

            result := -1;
            try
            angle1 := integer(directionAngle.Items[err]);
            len1 := integer(directionLen.Items[err]);

            angle2 := integer(directionAngle.Items[err+1]);
            len2 := integer(directionLen.Items[err+1]) - len1;

            result := angle1 * ((len1+len2) div len1) + angle2 * ((len1+len2) div len2);
            finally
            end;
        end;
    end;

    function ReplaceError(dir1,dir2, rep : TMouseDirection) : string;
    begin
        result := StringReplace(gesture, IntToStr(Integer(dir1)) +IntToStr(Integer(dir2)), IntToStr(Integer(rep)),[]);
    end;
var ang : integer;
begin
    ang := GetExtrapAngle(mdUp, mdNE);
    if ang <> -1 then begin
        if (ang - 270) < ((270+45) - ang) then begin
            suggest.Add(ReplaceError(mdUp, mdNE, mdUP));
            suggest.Add(ReplaceError(mdUp, mdNE, mdNE));
        end else begin
            suggest.Add(ReplaceError(mdUp, mdNE, mdNE));
            suggest.Add(ReplaceError(mdUp, mdNE, mdUP));
        end;
    end;
    ang := GetExtrapAngle(mdNE, mdUp);
    if ang <> -1 then begin
        if (ang - 270) < ((270+45) - ang) then begin
            suggest.Add(ReplaceError(mdNE, mdUp, mdUP));
            suggest.Add(ReplaceError(mdNE, mdUp, mdNE));
        end else begin
            suggest.Add(ReplaceError(mdNE, mdUp, mdNE));
            suggest.Add(ReplaceError(mdNE, mdUp, mdUP));
        end;
    end;

    ang := GetExtrapAngle(mdUp, mdNW);
    if ang <> -1 then begin
        if (270 - ang) < ((270-45) - ang) then begin
            suggest.Add(ReplaceError(mdUp, mdNW, mdUP));
            suggest.Add(ReplaceError(mdUp, mdNW, mdNW));
        end else begin
            suggest.Add(ReplaceError(mdUp, mdNW, mdNW));
            suggest.Add(ReplaceError(mdUp, mdNW, mdUP));
        end;
    end;
    ang := GetExtrapAngle(mdNW, mdUp);
    if ang <> -1 then begin
        if (270 - ang) < ((270-45) - ang) then begin
            suggest.Add(ReplaceError(mdNW, mdUp, mdUP));
            suggest.Add(ReplaceError(mdNW, mdUp, mdNW));
        end else begin
            suggest.Add(ReplaceError(mdNW, mdUp, mdNW));
            suggest.Add(ReplaceError(mdNW, mdUp, mdUP));
        end;
    end;

    ang := GetExtrapAngle(mdLeft, mdNW);
    if ang <> -1 then begin
        if (ang - 180) < ((180+45) - ang) then begin
            suggest.Add(ReplaceError(mdLeft, mdNW, mdLeft));
            suggest.Add(ReplaceError(mdLeft, mdNW, mdNW));
        end else begin
            suggest.Add(ReplaceError(mdLeft, mdNW, mdNW));
            suggest.Add(ReplaceError(mdLeft, mdNW, mdLeft));
        end;
    end;
    ang := GetExtrapAngle(mdNW, mdLeft);
    if ang <> -1 then begin
        if (ang - 180) < ((180+45) - ang) then begin
            suggest.Add(ReplaceError(mdNW, mdLeft, mdLeft));
            suggest.Add(ReplaceError(mdNW, mdLeft, mdNW));
        end else begin
            suggest.Add(ReplaceError(mdNW, mdLeft, mdNW));
            suggest.Add(ReplaceError(mdNW, mdLeft, mdLeft));
        end;
    end;

    ang := GetExtrapAngle(mdLeft, mdSW);
    if ang <> -1 then begin
        if (180 - ang) < ((180-45) - ang) then begin
            suggest.Add(ReplaceError(mdLeft, mdSW, mdLeft));
            suggest.Add(ReplaceError(mdLeft, mdSW, mdSW));
        end else begin
            suggest.Add(ReplaceError(mdLeft, mdSW, mdSW));
            suggest.Add(ReplaceError(mdLeft, mdSW, mdLeft));
        end;
    end;
    ang := GetExtrapAngle(mdSW, mdLeft);
    if ang <> -1 then begin
        if (180 - ang) < ((180-45) - ang) then begin
            suggest.Add(ReplaceError(mdSW, mdLeft, mdLeft));
            suggest.Add(ReplaceError(mdSW, mdLeft, mdSW));
        end else begin
            suggest.Add(ReplaceError(mdSW, mdLeft, mdSW));
            suggest.Add(ReplaceError(mdSW, mdLeft, mdLeft));
        end;
    end;


    ang := GetExtrapAngle(mdDown, mdSW);
    if ang <> -1 then begin
        if (90 - ang) < (ang - (180-45)) then begin
            suggest.Add(ReplaceError(mdDown, mdSW, mdDown));
            suggest.Add(ReplaceError(mdDown, mdSW, mdSW));
        end else begin
            suggest.Add(ReplaceError(mdDown, mdSW, mdSW));
            suggest.Add(ReplaceError(mdDown, mdSW, mdDown));
        end;
    end;
    ang := GetExtrapAngle(mdSW, mdDown);
    if ang <> -1 then begin
        if (90 - ang) < (ang - (180-45)) then begin
            suggest.Add(ReplaceError(mdSW, mdDown, mdDown));
            suggest.Add(ReplaceError(mdSW, mdDown, mdSW));
        end else begin
            suggest.Add(ReplaceError(mdSW, mdDown, mdSW));
            suggest.Add(ReplaceError(mdSW, mdDown, mdDown));
        end;
    end;


    ang := GetExtrapAngle(mdDown, mdSE);
    if ang <> -1 then begin
        if (ang - 90) < ((90-45)-ang) then begin
            suggest.Add(ReplaceError(mdDown, mdSE, mdDown));
            suggest.Add(ReplaceError(mdDown, mdSE, mdSE));
        end else begin
            suggest.Add(ReplaceError(mdDown, mdSE, mdSE));
            suggest.Add(ReplaceError(mdDown, mdSE, mdDown));
        end;
    end;
    ang := GetExtrapAngle(mdSE, mdDown);
    if ang <> -1 then begin
        if (ang - 90) < ((90-45)-ang) then begin
            suggest.Add(ReplaceError(mdSE, mdDown, mdDown));
            suggest.Add(ReplaceError(mdSE, mdDown, mdSE));
        end else begin
            suggest.Add(ReplaceError(mdSE, mdDown, mdSE));
            suggest.Add(ReplaceError(mdSE, mdDown, mdDown));
        end;
    end;

    ang := GetExtrapAngle(mdRight, mdSE);
    if ang <> -1 then begin
        if (ang) < (45-ang) then begin
            suggest.Add(ReplaceError(mdRight, mdSE, mdRight));
            suggest.Add(ReplaceError(mdRight, mdSE, mdSE));
        end else begin
            suggest.Add(ReplaceError(mdRight, mdSE, mdSE));
            suggest.Add(ReplaceError(mdRight, mdSE, mdRight));
        end;
    end;
    ang := GetExtrapAngle(mdSE, mdRight);
    if ang <> -1 then begin
        if (ang) < (45-ang) then begin
            suggest.Add(ReplaceError(mdSE, mdRight, mdRight));
            suggest.Add(ReplaceError(mdSE, mdRight, mdSE));
        end else begin
            suggest.Add(ReplaceError(mdSE, mdRight, mdSE));
            suggest.Add(ReplaceError(mdSE, mdRight, mdRight));
        end;
    end;

    ang := GetExtrapAngle(mdRight, mdNE);
    if ang <> -1 then begin
        if (359 - ang) < ((259-45)-ang) then begin
            suggest.Add(ReplaceError(mdRight, mdNE, mdRight));
            suggest.Add(ReplaceError(mdRight, mdNE, mdNE));
        end else begin
            suggest.Add(ReplaceError(mdRight, mdNE, mdNE));
            suggest.Add(ReplaceError(mdRight, mdNE, mdRight));
        end;
    end;
    ang := GetExtrapAngle(mdNE, mdRight);
    if ang <> -1 then begin
        if (359 - ang) < ((259-45)-ang) then begin
            suggest.Add(ReplaceError(mdNE, mdRight, mdRight));
            suggest.Add(ReplaceError(mdNE, mdRight, mdNE));
        end else begin
            suggest.Add(ReplaceError(mdNE, mdRight, mdNE));
            suggest.Add(ReplaceError(mdNE, mdRight, mdRight));
        end;
    end;

end;

function TFrmMouseMonitor.IsBetweenNeighbors(dir1, dir2: TMouseDirection;
  angle: integer): boolean;
    function DirectionToAngle(dir : TMouseDirection) : integer;
    begin
        case dir of
            mdUp: result := UP_A;
            mdDown: result := DOWN_A;
            mdLeft: result := LEFT_A;
            mdRight: result := RIGHT_A;
            mdNE: result := NE_A;
            mdNW: result := NW_A;
            mdSE: result := SE_A;
            mdSW: result := SW_A;
        end;
    end;
    function Neighbors(dir1, dir2 : TMouseDirection) : boolean;
    begin
        result := false;
        if (dir1 = mdUp) and (dir2 in [mdNE,mdNW])  then begin
            result := true;
        end else if (dir1 = mdNE) and (dir2 in [mdUP,mdRight]) then begin
            result := true;
        end else if (dir1 = mdRight) and (dir2 in [mdNE,mdSW]) then begin
            result := true;
        end else if (dir1 = mdSE) and (dir2 in [mdRight,mdDown]) then begin
            result := true;
        end else if (dir1 = mdDown) and (dir2 in [mdSE,mdSW]) then begin
            result := true;
        end else if (dir1 = mdSW) and (dir2 in [mdDown,mdLeft]) then begin
            result := true;
        end else if (dir1 = mdLeft) and (dir2 in [mdSW,mdNW]) then begin
            result := true;
        end else if (dir1 = mdNW) and (dir2 in [mdLeft,mdUp]) then begin
            result := true;
        end;

    end;
begin
    result := Neighbors(dir1,dir2) and
        (
            (DirectionToAngle(dir1) >= angle) and (DirectionToAngle(dir2) <= angle)
            or
            (DirectionToAngle(dir2) >= angle) and (DirectionToAngle(dir1) <= angle)
        );
end;

function TFrmMouseMonitor.IsDiagonal(md: TMouseDirection): boolean;
begin
    result := md in [mdNE, mdSE, mdNW, mdSW];
end;

function TFrmMouseMonitor.IsFullScreen: boolean;
var h  : Thandle;
    r : TRect;
    m : TMonitor;
    name : string;
    wp : tagWINDOWPLACEMENT;
    maxed : boolean;
begin

    result := true;
    h := Windows.GetForegroundWindow;
    name := WindowHandleToEXEName(h);
    if name = 'explorer.exe' then begin
        // assume it's the Desktop since there's no associated EXE
        result := false;
        EXIT;
    end;


    if GetWindowRect(h,r) then begin
        m := screen.MonitorFromRect(r);
        if m = nil then begin
            // Major Kludge - showing a window somehow resets the
            // monitor list after a screen is detached from the OS
            self.Width := 10;
            self.Height := 10;


            Windows.ShowWindow(self.handle, SW_MINIMIZE);
            self.show;
            self.hide;
            m := screen.MonitorFromWindow(h);
            if m = nil then begin
                UnitMisc.AppendLog('MonitorFromRect failed', true);
                exit;
            end;
        end;

        wp.length := sizeof(wp);
        maxed := true;
        if Windows.GetWindowPlacement(h,@wp) then begin
            maxed := wp.showCmd = SW_SHOWMAXIMIZED;
        end;
        //UnitMisc.AppendLog('RightClickTrigger test ' + IntToStr(wp.showCmd));
        if ((r.Bottom-r.Top) >= (m.BoundsRect.Bottom-m.BoundsRect.top)) and
            ((r.Right-r.left) >= (m.boundsrect.Right-m.boundsrect.Left)) and
            not (maxed)
        then begin
            EXIT;
        end;
        result := false;
    end;

end;

function TFrmMouseMonitor.IsNotDiagonal(md: TMouseDirection): boolean;
begin
     result := not IsDiagonal(md);
end;

procedure TFrmMouseMonitor.SetMonitorEnabled(enable: boolean);
begin
    timMouse.Enabled := enable;
    disabled := not enable;
end;

procedure TFrmMouseMonitor.SetHelpPause(const Value: integer);
begin
    timPaused.Interval := value;
    self.helpPause := value;
end;

procedure TFrmMouseMonitor.timMouseTimer(Sender: TObject);

    // utility functions
    function ExpAvg(last, current : integer; lastbias : single) : integer;
    begin
        if last = 0 then last := current;

        result := trunc( (last*lastbias) + (current*(1-lastbias)) );
    end;

    procedure NewLine(i : integer; name : string);
    begin
        AppendLog(name + ':' + IntToStr(i));
    end;
    procedure ClearDirectionData;
    begin
        xavg := 0;
        yavg := 0;
        upavg := 0;
        downavg := 0;
        leftavg := 0;
        rightavg := 0;
        angleavg := 0;
        upcount := 0;
        downcount := 0;
        leftcount := 0;
        rightcount := 0;
        NEcount := 0;
        SEcount := 0;
        NWcount := 0;
        SWcount := 0;
        NEavg := 0;
        SEavg := 0;
        NWavg := 0;
        SWavg := 0;
        lastangle := 0;
        xSwapped := false;
        ySwapped := false;
        
    end;

    function PolarAngle(startX,startY, endX,endY : integer) : integer;
    begin
        // R=Sqrt(x2+y2);
        // Theta=ArcTan(Y/X);

        // move the origin to the starting point
        endX := endX - startX;
        endY := endY - startY;
        result := 0;
        try
            // 0 to 360 convert
            if (endX > 0) and (endY >= 0) then
                result := trunc(ArcTan(endY/endX) * (180 / Pi))
            else if (endX > 0) and (endY < 0) then
                result := trunc((ArcTan(endY/endX) + 2*Pi) * (180 / Pi))
            else if (endX < 0) then
                result := trunc((ArcTan(endY/endX) + Pi) * (180 / Pi))
            else if (endX = 0) and (endY > 0) then
                result := trunc((Pi/2) * (180 / Pi))
            else if (endX = 0) and (endY > 0) then
                result := trunc(3*(Pi/2) * (180 / Pi));

        except
        end;

    end;


    function GetCurrentMouseAngle : integer;
    begin
        result := -1;
        if storedx*storedy = 0 then EXIT;


        // ignore mouse angles for movements smaller than a fraction of the
        // length threshold
        {if abs(storedx-mouse.CursorPos.X) < (3) then EXIT;
        if abs(storedy-mouse.CursorPos.y) < (3) then EXIT;}

        //m := Screen.MonitorFromPoint(mouse.CursorPos);
        result := PolarAngle(storedX,storedY, mouse.CursorPos.X, mouse.CursorPos.Y);
    end;


    procedure CalculateDeltaAndReverses;
    var M : Tmonitor;
    begin
        // using the change in X,Y directions, create a resolution
        // independant acclerated average of the X,Y deltas
        if lastx = 0 then begin
            lastx := Mouse.CursorPos.x;
        end;
        if lasty = 0 then begin
            lasty := Mouse.CursorPos.y;
        end;
        xdelta := Mouse.CursorPos.X - lastx;
        ydelta := Mouse.CursorPos.Y - lasty;

        m := Screen.MonitorFromPoint(mouse.CursorPos);
        if m = nil then begin
            UnitMisc.AppendLog('MonitorFromRect failed', true);
            exit;
        end;
        // change the pixels into a percenct as a whole number (0 to 1000)
        if (xdelta <> 0) and (ydelta <> 0) then begin
            xavg := ExpAvg(trunc((xdelta/m.Width)*1800), xavg, deltaBias);
            yavg := ExpAvg(trunc((ydelta/m.Height)*1800), yavg, deltaBias);
        end;
        // detect a reverse in horizontal direction
        xSwapped := false;
        if xavg > 0 then begin
            if (xbinary = 0) or (xbinary = -1) then
                xSwapped := true;
            xbinary := 1;
        end else if xavg < 0 then begin
            if (xbinary = 0) or (xbinary = 1) then
                xSwapped := true;
            xbinary := -1;
        end;
        ySwapped := false;
        if yavg > 0 then begin
            if (ybinary = 0) or (ybinary = -1) then
                ySwapped := true;
            ybinary := 1;
        end else if yavg < 0 then begin
            if (ybinary = 0) or (ybinary = 1) then
                ySwapped := true;
            ybinary := -1;
        end;

    end;

    procedure CountThresholdCrosses;
    begin
        // Increase a direction counter when the threshold has been reached
        vert := mdNone;
        horz := mdNone;
        if abs(xavg) > xThreshold then begin
            if xavg > 0 then begin
                inc(rightcount);
                leftcount := 0;
                horz := mdRight;
            end else begin
                inc(leftcount);
                rightcount := 0;
                horz := mdLeft;
            end;
        end;
        if abs(yavg) > yThreshold then begin
            if yavg < 0 then begin
                inc(upcount);
                downcount := 0;
                vert := mdUp;
            end else begin
                inc(downcount);
                upcount := 0;
                vert := mdDown;
            end;
        end;

        eastQuad := mdNone;
        westQuad := mdNone;
        if (abs(xavg) > xThreshold) and
            (abs(yavg) > yThreshold) then begin
            if xavg > 0 then begin // east
                if yavg > 0 then begin // south
                    inc(SEcount);
                    NECount := 0;
                    eastquad := mdSE;
                end else begin // north
                    inc(NEcount);
                    SEcount := 0;
                    eastQuad := mdNE;
                end;
            end else begin // west
                if yavg > 0 then begin // south
                    inc(SWcount);
                    NWCount := 0;
                    westQuad := mdSW;
                end else begin // north
                    inc(NWcount);
                    SWcount := 0;
                    westQuad := mdNW;
                end;
            end;
        end;

    end;
    procedure UpdateDirectionAverages;
    begin
        // accelerating count of the times each direction is above the threshold
        if vert = mdUp then begin
            upavg := ExpAvg(upcount, upavg, directionBias);
            downavg := ExpAvg(0,downavg, directionBias);
        end else if vert = mdDown then begin
            upavg := ExpAvg(0, upavg, directionBias);
            downavg := ExpAvg(downcount, downavg, directionBias );
        end else begin
            upavg := ExpAvg(0, upavg, directionBias);
            downavg := ExpAvg(0, downavg,directionBias);
        end;
        if horz = mdLeft then begin
            leftavg := ExpAvg(leftcount, leftavg, directionBias );
            rightavg := ExpAvg(0,rightavg, directionBias);
        end else if horz = mdRight then begin
            leftavg := ExpAvg(0,leftavg, directionBias);
            rightavg := ExpAvg(rightcount, rightavg, directionBias );
        end else begin
            leftavg := ExpAvg(0, leftavg, directionBias);
            rightavg := ExpAvg(0, rightavg, directionBias);
        end;

        if Self.EnableDiagonals then  begin
            if eastQuad = mdNE then begin
                NEavg := ExpAvg(NEcount, NEavg, directionBias);
                SEavg := ExpAvg(0, SEavg, directionBias);
            end else if eastQuad = mdSE then begin
                NEavg := ExpAvg(0, NEavg, directionBias);
                SEavg := ExpAvg(SEcount, SEavg, directionBias);
            end else begin
                NEavg := ExpAvg(0, NEavg, directionBias);
                SEavg := ExpAvg(0, SEavg, directionBias);
            end;

            if westQuad = mdNW then begin
                NWavg := ExpAvg(NWcount, NWavg, directionBias);
                SWavg := ExpAvg(0, SWavg, directionBias);
            end else if westQuad = mdSW then begin
                NWavg := ExpAvg(0, NWavg, directionBias);
                SWavg := ExpAvg(SWcount, SWavg, directionBias);
            end else begin
                NWavg := ExpAvg(0, NWavg, directionBias);
                SWavg := ExpAvg(0, SWavg, directionBias);
            end;
        end;



    end;
    procedure UpdateVertHorzPercent;
    begin
    end;

    procedure CorrectErrorDirection(EnableDeadzone : boolean = true);
    var a, b : integer;
    const   DIAG_PLUSMINUS = 20;
            HALF_DEADZONE = 4;
        function SubAngle(a : integer; subtract : integer) : integer;
        begin
            result := a - subtract;
            if result < 0 then begin
                result := 360 + result;
            end;
        end;
        function PlusAngle(a : integer; add : integer) : integer;
        begin
            result := a + add;
            if result > 360 then begin
                result := result - 360;
            end;
        end;
    begin
        a := GetCurrentMouseAngle;
        if a = -1 then begin
            EXIT;
        end;



            // RIGHT/EAST is a special case where it's wraps between > 360 and < 0
            if (a >= SubAngle(RIGHT_A,45)) or (a < PlusAngle(RIGHT_A,45)) then begin         // right
                horzDirection := mdRight;
            end else  if (a >= SubAngle(DOWN_A,45)) and (a < PlusAngle(DOWN_A,45)) then begin // down
                vertDirection := mdDown;
            end else if (a >= SubANgle(LEFT_A,45)) and ( a < PlusAngle(LEFT_A,45)) then begin // left
                horzDirection := mdLeft;
            end else if (a >= SubANgle(UP_A,45)) and (a < PlusAngle(UP_A,45)) then begin      // up
                vertDirection := mdUp;
            end;


            if self.EnableDiagonals then begin
                eastDirection := mdNone;
                westDirection := mdNone;
                if (a >= SubAngle(NE_A,DIAG_PLUSMINUS)) and (a < PlusAngle(NE_A,DIAG_PLUSMINUS)) then begin
                    eastDirection := mdNE;
                    horzDirection := mdRight;
                    vertDirection := mdUp;
                end else if (a >= SubAngle(SE_A,DIAG_PLUSMINUS)) and (a < PlusAngle(SE_A,DIAG_PLUSMINUS)) then begin
                    eastDirection := mdSE;
                    horzDirection := mdRight;
                    vertDirection := mdDown;
                end else if (a >= SubAngle(NW_A,DIAG_PLUSMINUS)) and (a < PlusAngle(NW_A,DIAG_PLUSMINUS)) then begin
                    westDirection := mdNW;
                    horzDirection := mdLeft;
                    vertDirection := mdUp;
                end else if (a >= SubAngle(SW_A,DIAG_PLUSMINUS)) and (a < PlusAngle(SW_A,DIAG_PLUSMINUS)) then begin
                    westDirection := mdSW;
                    horzDirection := mdLeft;
                    vertDirection := mdDown;
                end;

                // detect deadzones
                // within a few degrees of the start of diagonal cutoff
                if EnableDeadzone then begin
                    b := a mod 90;
                    if (b > (DIAG_PLUSMINUS-HALF_DEADZONE)) and (b< (DIAG_PLUSMINUS+HALF_DEADZONE)) then begin
                        westDirection := mdNone;
                        eastDirection := mdNone;
                        horzDirection := mdNone;
                        vertDirection := mdNOne;
                        AppendLog('dead zone');
                    end;
                    b := 90 - b;
                    if (b > (DIAG_PLUSMINUS-HALF_DEADZONE)) and (b< (DIAG_PLUSMINUS+HALF_DEADZONE)) then begin
                        westDirection := mdNone;
                        eastDirection := mdNone;
                        horzDirection := mdNone;
                        vertDirection := mdNOne;
                        AppendLog('dead zone');
                    end;
                end;



            end;

        txtPercent.Caption := IntToSTr(a);
        self.lastangle := a;

        angleavg := ExpAvg(angleavg, lastangle, 0.35);


    end;

    procedure DetermineMainDirection; forward;
    procedure DetectLongDeadZoneMovement;
    var multiple : double;
    begin
        multiple := 1.5;

        if(storedx <> 0) and (storedy <> 0) then
            if ((abs(storedx - mouse.cursorpos.X) > trunc(Self.XPixelThreshold *  multiple)) or
                (abs(storedy - mouse.cursorpos.y) > trunc(Self.YPixelThreshold *  multiple)))
                and (detectedDirection = mdNone) then begin
                CorrectErrorDirection(False);
                //AppendLog(' - ' + ToString(horzDirection) + ToSTring(vertDirection));
                DetermineMainDirection;
                //AppendLog(' - ' + ToSTring(detectedDirection));
                AppendLog('long dead zone override ' + toString(detectedDirection));
                
            end;
    end;


    procedure DetermineXYDirections;
    var a : integer;
    begin
        // main direction are based off of direction averages
        vertdirection := mdNone;
        horzDirection := mdNone;
        eastDirection := mdNone;
        westDirection := mdNone;
        if (upavg > directionThreshold) or (downavg > directionThreshold) or
            (leftavg > directionThreshold) or (rightavg > directionThreshold) or
            (NEavg > directionThreshold) or (NWavg > directionThreshold) or
                (SEavg > directionThreshold) or (SWavg > directionThreshold) then begin
            CorrectErrorDirection;

        end;
    end;

    function AngleFromOppAdj(opp, adj : integer) : integer;
    begin
        result := 0;
        if (opp*adj) = 0 then EXIT;

        try
            result := trunc( arctan(opp/adj)  * (180 / PI) );
        except
        end;
    end;

    function LenFromPoints(x1,y1,x2,y2 : integer) : integer;
    var miny, maxy : integer;
        minx, maxx : integer;
        a,b : integer;
    begin
        // logicall rotate to place "opposite" leg as Y
        if abs(y1-y2) < abs(x1-x2) then begin
            miny := min(y1,y2);
            maxy := max(y1,y2);

            minx := min(x1,x2);
            maxx := max(x1,x2);
        end else begin
            miny := min(x1,x2);
            maxy := max(x1,x2);

            minx := min(y1,y2);
            maxx := max(y1,y2);
        end;

        if (x1 = 0) or (y1=0) or (x1=0) or (y1=0) then
        begin
            result := 0;
        end else
        begin

        b := maxx - minx; // adjacent
        a := maxy - miny; // opposite
        result := trunc( sqrt(a*a + b*b) );
        end;
    end;



    procedure DetermineMainDirection;
    begin
        {if self.EnableDiagonals then begin
            if peravg >= 1 then begin
                detectedDirection := mdNone;
            end;
        end;}
        if not (vertDirection = mdNone) and not (horzDirection = mdNone) then begin
            // translate movment delata into a percent based on screen ratio
            if self.EnableDiagonals then begin
                if eastDirection <> mdNone then begin
                    detectedDirection := eastDirection; // SE NE
                end else begin
                    detectedDirection := westDirection; // SW NW
                end;
            end else begin
                if abs(xavg)>abs(yavg) then begin
                    detectedDirection := horzDirection;
                end else begin
                    detectedDirection := vertDirection;
                end;
            end;
        end else if (vertDirection = mdNone) then begin
            detectedDirection := horzDirection;

        end else begin
            detectedDirection := vertDirection;
        end;

       // AppendLog(' main - ' + ToString(detectedDirection));
    end;
    procedure UpdateDebugInfo;
    begin
        txtXavg.caption := IntToSTr(xavg);
        txtYavg.caption := IntToStr(yavg);
        txtup.caption := IntToStr(upavg);
        txtdown.caption := IntToStr(downavg);
        txtleft.caption := IntToStr(leftavg);
        txtright.caption := IntToStr(rightavg);

        txtStartX.caption :=  IntToStr(storedX);
        txtStartY.caption := IntToStr(storedY);
        txtmX.caption := IntToStr(Mouse.CursorPos.X);
        txtmY.caption := IntToStr(Mouse.CursorPos.Y);
        txtPercent.Caption := IntToStr(angleavg);
        case detectedDirection  of
            mdUp: txtDirection.Caption := 'Up';
            mdDown: txtDirection.Caption := 'Down';
            mdLeft: txtDirection.Caption := 'Left';
            mdRight: txtDirection.Caption := 'Right';
            mdNone: txtDirection.Caption := 'None';
        end;
        //AppendLog(txtDirection.caption);

    end;


    procedure UpdateDebugLog;
    begin
        NewLine(timeheld, 'timeheld');
        NewLine(upcount, 'upcount');
        NewLine(downcount, 'downcount');
        NewLine(leftcount, 'leftcount');
        NewLine(rightcount, 'rightcount');
    end;
    procedure CheckForTooltip(adddirection : TMouseDirection);
    begin
        if LastTooltipText = '' then begin
            LastTooltipText := ToString(adddirection);
            case adddirection of
                mdUp: tt.ShowTooltip(self.imgUp.Picture,point(Mouse.CursorPos.X - 90, mouse.cursorpos.y+5),'');
                mdDown: tt.ShowTooltip(self.imgDown.Picture,point(Mouse.CursorPos.X - 90, mouse.cursorpos.y-40),'');
                mdLeft: tt.ShowTooltip(self.imgLeft.Picture,point(Mouse.CursorPos.X - 90, mouse.cursorpos.y+30),'');
                mdRight: tt.ShowTooltip(self.imgRight.Picture,point(Mouse.CursorPos.X - 90, mouse.cursorpos.y+30),'');
                mdNE: tt.ShowTooltip(self.imgNE.Picture,point(Mouse.CursorPos.X - 90, mouse.cursorpos.y+30),'');
                mdNW: tt.ShowTooltip(self.imgNW.Picture,point(Mouse.CursorPos.X - 90, mouse.cursorpos.y+30),'');
                mdSE: tt.ShowTooltip(self.imgSE.Picture,point(Mouse.CursorPos.X - 90, mouse.cursorpos.y+30),'');
                mdSW: tt.ShowTooltip(self.imgSW.Picture,point(Mouse.CursorPos.X - 90, mouse.cursorpos.y+30),'');
            end;

        end else if LastTooltipText <> ToString(adddirection) then begin
            LastTooltipText := ToString(adddirection);
            tt.CloseTooltip(false);
            case adddirection of
                mdUp: tt.ShowTooltip(self.imgUp.Picture,point(Tt.left,tt.top),'');
                mdDown: tt.ShowTooltip(self.imgDown.Picture,point(Tt.left,tt.top),'');
                mdLeft: tt.ShowTooltip(self.imgLeft.Picture,point(Tt.left,tt.top),'');
                mdRight: tt.ShowTooltip(self.imgRight.Picture,point(Tt.left,tt.top),'');
                mdNE: tt.ShowTooltip(self.imgNE.Picture,point(Tt.left,tt.top),'');
                mdNW: tt.ShowTooltip(self.imgNW.Picture,point(Tt.left,tt.top),'');
                mdSE: tt.ShowTooltip(self.imgSE.Picture,point(Tt.left,tt.top),'');
                mdSW: tt.ShowTooltip(self.imgSW.Picture,point(Tt.left,tt.top),'');
            end;
        end;
    end;

    function DetectDirectionChanges : boolean;
    begin
        result := false;
        // very first interp
        {if (firstdirection = mdNone) and interpreted then begin
            storedX := mouse.CursorPos.x;
            storedY := mouse.CursorPos.y;
            lastX := storedX;
            lastY := storedY;
            AppendLog(' first direction empty x,y ' + IntToStr(lastx) + ',' + IntToStr(lasty));
            ClearDirectionData;
        end;
        //}
        if (storedx = 0) and (storedy = 0) then begin
            storedX := lastx;
            storedY := lasty;
            //lastX := storedX;
            //lastY := storedY;
            AppendLog(' no stored data ' + IntToStr(lastx) + ',' + IntToStr(lasty));
            ClearDirectionData;
        end else if xSwapped or ySwapped then begin
            if firstdirection = mdNone then begin
                if storedx*storedy <> 0 then begin
                    if (abs(storedy-lasty) <= YPixelThreshold) and
                        (abs(storedx-lastx) <= XPixelThreshold)
                    then begin
                        EXIT;
                    end;
                end;
            end;
            
            storedX := lastx;
            storedY := lasty;
            //lastX := storedX;
            //lastY := storedY;
            AppendLog(' direction reversed x,y ' + IntToStr(lastx) + ',' + IntToStr(lasty));
            ClearDirectionData;
        end;

        if DetectedDirection <> mdNone then begin

            // store when a diagonal to normal is detected, so the rejector can get rid of
            // small errors at the tail end of the direction
            if IsDiagonal(lastDetectedDirection) and IsNotDiagonal(detectedDirection)   then begin
                if ((lastDetectedDirection = mdDown) and (detectedDirection in [mdSE, mdSW])) or
                    ((lastDetectedDirection = mdUP) and (detectedDirection in [mdNE, mdSW])) or
                    ((lastDetectedDirection = mdLeft) and (detectedDirection in [mdSW, mdNW])) or
                    ((lastDetectedDirection = mdRight) and (detectedDirection in [mdSE, mdNE]))
                then begin
                    AppendLog(' LastDetect ' + ToString(lastDetectedDirection));
                    AppendLog(' Direction Change ' + ToString(detectedDirection));
                    storedX := lastx;
                    storedY := lasty;
                    AppendLog(' Point stored for tail errors x,y ' + IntToStr(storedX) + ',' + IntToStr(storedy));

                    ClearDirectionData;

                    result := true;
                end;
            end;
            {
            // mouse is moving in a new direction from the last added direction
            // but the last movement was in the same direction as the last added
            if (LastAddActual <> mdNone) and
                (LastAddActual <> detectedDirection) and
                (lastDetectedDirection = lastAddDirection) and
                (lastDetectedDirection <> mdNone)
               then begin
                storedX := lastx;
                storedY := lasty;
                AppendLog(' Point stored x,y ' + IntToStr(storedX) + ',' + IntToStr(storedY));
                AppendLog(' Mouse x,y ' + IntToStr(mouse.CursorPos.x) + ',' + IntToStr(Mouse.CursorPos.y));

                AppendLog(' LastAddActual ' + ToString(LastAddActual));
                AppendLog(' Direction Change ' + ToString(detectedDirection));
                ClearDirectionData;

                result := true;
            end;
            //}
        end;
    end;
    function DirectionPixelCheck(d : TMouseDirection) : TMouseDirection;
    begin
        if (d = mdNone) then begin
            //AppendLog(' detected NONE');
        end;
        result := d;
        if d in [mdUP,mdDown] then begin
            if abs(storedY - Mouse.CursorPos.Y) < YPixelThreshold  then begin
                result := mdNone;
                AppendLog(' Pixel Failed ');
            end;
        end else if d in [mdLeft, mdRight] then begin
            if abs(storedX - Mouse.CursorPos.X) < XPixelThreshold  then begin
                result := mdNone;
                AppendLog(' Pixel Failed ');
            end;
        end else if d in [mdNE, mdNW, mdSE, mdSW] then begin
            if ((abs(storedY - Mouse.CursorPos.Y) +  abs(storedX - Mouse.CursorPos.X)) < trunc(YPixelThreshold * 1.3)) 
                then begin
                result := mdNone;
                AppendLog(' Diagonal Pixel Failed ');
            end;
        end;
    end;

    function SteepAngleCheck(add : TMouseDirection) : TMouseDirection;
    var angle : integer;
    begin
        result := add;
        if self.EnableDiagonals then begin
        end;
    end;
    function DetectErrorCombination(m : TMouseDirection) : TMouseDirection;
    begin
        result := m;
        if (LastAddActual <> mdNone) then begin
            case LastAddActual  of
                mdUp:  if m in [mdNE,mdNW] then result := mdNone;
                mdDown: if m in [mdSW, mdSE] then result := mdNone;
                mdLeft: if m in [mdSW, mdNW] then result := mdNone;
                mdRight: if m in [mdSE, mdNE] then result := mdNone;
                mdNE: if m in [mdRight, mdUp] then result := mdNone;
                mdNW: if m in [mdLeft, mdUp] then result := mdNone;
                mdSE: if m in [mdRight, mdDown] then result := mdNone;
                mdSW: if m in [mdLeft, mdDown] then result := mdNone;
            end;
            if LastAddActual = mdNone then begin
                AppendLog(' error removed ' + ToSTring(LastAddActual)); 
            end;
            
        end;

        if result <> m then begin
            storedX := lastx;
            storedY := lasty;
            //lastX := storedX;
            //lastY := storedY;
            AppendLog(' error combo x,y ' + IntToStr(lastx) + ',' + IntToStr(lasty));
            ClearDirectionData;
        end;
    end;
    procedure DetectSlowLongMovement;
    begin
        // if the speed threshold hasn't been crossed, but the mouse has
        // move over the pixel threshold, force direction detection
        if (abs(storedY-mouse.CursorPos.Y) > YPixelThreshold) or
            (abs(storedX-mouse.CursorPos.X) > XPixelThreshold) then begin
            //AppendLog('slow movement');
            CorrectErrorDirection;
            //AppendLog(ToString(horzDirection) + ToString(vertDirection));

            DetermineMainDirection;
        end;

    end;

    procedure InitVars;
    begin
        AppendLog(' - Init - ');
        xdelta := 0;
        ydelta := 0;
        xavg := 0;
        yavg := 0;
        lastx := mouse.CursorPos.x;
        lasty := mouse.CursorPos.y;
        timeheld := 0;
        leftcount := 0;
        rightcount := 0;
        upcount := 0;
        downcount := 0;
        upavg := 0;
        rightavg := 0;
        leftavg := 0;
        rightavg := 0;
        peravgmax := 0;
        angleavg := 0;

        startingx := lastx;
        startingy := lasty;

        NEcount := 0;
        NWcount := 0;
        SEcount := 0;
        SWcount := 0;
        NEavg := 0;
        SEavg := 0;
        NWavg := 0;
        SWavg := 0;
        lastangle := 0;

        xbinary := 0;
        ybinary := 0;
        xSwapped := false;
        ySwapped := false;
        StoredX := 0;
        SToredY := 0;
        
        detectedDirection := mdNone;
        txtDirection.caption := '';
        firstdirection := mdNone;
        LastTooltipText := '';
        LastAddActual := mdNone;
        LastChangeDirection := mdNone;
        // NOTE: The "first" case pretends we're starting over after
        // already interpreting a gesture
        interpreted := true;
    end;
var i : integer;
    md, lastmd : TMouseDirection;

    //storedDirection : TMouseDirection;
    addDirection : TMouseDirection;
    gesture : string;
    h : Thandle;
    angle : integer;
    disabled : boolean;
begin
    if abort then begin
        abort := false;
        FrmRatTrap.Visible := false;
        InitVars;
        EXIT;
    end;
    if disabled then EXIT;
    
    if KeyboardQuery.IsClicked(RightButton)   then begin
        timMouse.Interval := 40;

        FrmScratchPad.ReportMouseButton(true);
        if interpreted then begin
            memolog.Lines.Clear;
            interpreted := false;
            directionlist.Clear;
            directionLen.clear;
            directionAngle.clear;
            InitVars;

            fullscreen := IsFullScreen;
            disabled := FrmGestures.ProgramIsDisabled;
            AppendLog('--------' + BoolToSTr(fullscreen));
        end else begin
            //fullscreen := false;
        end;
        if fullscreen then EXIT;
        if disabled then EXIT;
        // ignore mouse strokes on the gesture list window
        //h := Windows.GetForegroundWindow;
        //if h = FrmGestures.Handle then EXIT;



        timMouse.Enabled := false;

        try
            CalculateDeltaAndReverses;
            CountThresholdCrosses;
            UpdateDirectionAverages;
            UpdateVertHorzPercent;

            DetermineXYDirections;

            DetermineMainDirection;
            if DetectDirectionChanges then begin

                CorrectErrorDirection;
                DetermineMainDirection;
            end;
            DetectSlowLongMovement;
            DetectLongDeadZoneMovement;


            UpdateDebugInfo;

            addDirection := DirectionPixelCheck(detectedDirection);
            addDirection := SteepAngleCheck(addDirection);
            addDirection := DetectErrorCombination(AddDirection);

            if (Abs(mouse.CursorPos.X - startingx)> XPixelThreshold) or
                (Abs(mouse.CursorPos.Y - startingy)> YPixelThreshold) then begin
                if EnableDrawing and ((xdelta <>0) or (ydelta<>0)) then begin
                    if FrmRatTrap.Visible = False then begin
                        FrmRatTrap.ReportMouse(point(startingx, startingy));
                    end;
                    if FrmRatTrap.Visible = False then FrmRatTrap.Visible := True;
                    FrmRatTrap.ReportMouse(mouse.CursorPos);
                end;
            end;



            if addDirection <> mdNone then begin
                // remove last direction if it's a duplicate
                if directionlist.count > 0 then begin
                    if lastAddDirection = addDirection then begin
                        //AppendLog(' duplicate delete ');
                        directionlist.Delete(directionlist.count-1);
                        directionLen.Delete(directionLen.count-1);
                        directionAngle.Delete(directionAngle.count-1);
                    end else begin
                        FrmRatTrap.ReportDirectionChange;
                    end;
                end;
                LastAddActual := AddDirection;
                directionlist.Add(pointer(addDirection));
                directionLen.Add(pointer( LenFromPoints(mouse.CursorPos.x,mouse.CursorPos.y, storedx,storedy) ));
                directionAngle.Add( Pointer((PolarAngle(storedx,storedY, mouse.CursorPos.x,mouse.CursorPos.y))) );
                CheckForTooltip(addDirection);
                AppendLog('ADDING ' + tostring(addDirection)+
                                  '     --len- ' + IntToStr( Integer(directionLen.Items[directionLen.count-1] )));

                FrmScratchPad.ReportPoint(mouse.CursorPos);
            end else if (addDirection = mdNone) then begin
            end;


            lastDetectedDirection := detectedDirection;
            lastAddDirection := addDirection;

            // pause detection resets
            if (firstdirection = mdNone) and (addDirection <> mdNone) then begin
                firstdirection := addDirection;
                timPaused.Enabled := false;

                //self.ClosePopups;
                //FrmScratchPad.RemoveSelfFromFocus();

                if self.EnableHelpPause  then begin
                    timPaused.Enabled := true;
                end;
            end;
            if (firstdirection <> mdNone) and (addDirection <> firstDirection) then
                    timPaused.Enabled := false;


            // store for data for next run
            lastx := Mouse.CursorPos.X;
            lasty := Mouse.CursorPos.Y;


            timeheld := timeheld + (Windows.GetTickCount - lastClock);
            lastClock := Windows.GetTickCount;
            interpreted := false;
        finally
            if not disabled then
                timMouse.Enabled := true;
        end;
    end else begin
        FrmRatTrap.Visible := False;
        fullscreen := false;

        timMouse.Interval := 70;
        timMouse.enabled := false;
        timPaused.Enabled := false;
        FrmScratchPad.ReportMouseButton(false);
        //FrmLines.ReportMouseButton(false);
        tt.CloseTooltip(true);
        try

        fullscreen := false;
        if not interpreted then begin

            //KeyboardQuery.ClearClick(RightButton);
            lastmd := mdnone;
            AppendLog('int-------');
            gesture := '';
            for i := 0 to directionlist.count - 1 do begin
                md := TMouseDirection(Integer(directionlist.items[i]));
                if (md <> lastmd) then  begin
                    if md <> mdNone then begin
                        gesture := gesture + IntToStr(Integer(md));
                    end;
                    case md of

                        mdUp: begin
                            AppendLog('Direction UP');
                        end;
                        mdDown: AppendLog('Direction DOWN');
                        mdLeft: AppendLog('Direction LEFT');
                        mdRight: AppendLog('Direction RIGHT');
                        mdNE:  AppendLog('Direction NE');
                        mdNW: AppendLog('Direction NW');
                        mdSE: AppendLog('Direction SE');
                        mdSW: AppendLog('Direction SW');

                        //mdNone: AppendLog('Direction ');
                    end;
                    AppendLog( '    len - ' + IntToStr( Integer(directionLen.Items[i]) ) );
                    AppendLog( '    Angle - (' + IntToStr( Integer(directionAngle.Items[i]) ) + ')');

                end;
                if md <> mdNone then
                    lastmd := md;

            end;
            interpreted := true;
            if gesture <> '' then  begin
                //ClosePopups;
                if self.EnableDiagonals  then begin
                    FrmGestures.ReportGesturesDiagonal(gesture);
                end else begin
                    FrmGestures.ReportGestures(gesture);
                end;
                //ClosePopups;
            end;

        end;
        lastClock := Windows.GetTickCount;
        finally
            if not disabled then
                timMouse.Enabled := true;
        end;
    end;
end;

procedure TFrmMouseMonitor.timPausedTimer(Sender: TObject);
begin
    tt.CloseTooltip(false); 
    timPaused.Enabled := false;
    AppendLog(' Pause Trigger ');
    //self.AbortInterpret;
    //KeyboardQuery.ClearClick(RightButton);
    FrmGestures.ReportFirstDirection(self.firstdirection);
end;

function TFrmMouseMonitor.ToString(md: TMouseDirection): string;
begin
    case md of
        mdUp: result := 'Up';
        mdDown: result := 'Down';
        mdLeft: result := 'Left';
        mdRight: result := 'Right';
        mdNE: result := 'NE';
        mdNW: result := 'NW';
        mdSE: result := 'SE';
        mdSW: result := 'SW';
        mdNone: result := 'None';
    end;
end;

end.
